import numpy as np
cimport numpy as np
cimport cython

# don't use np.sqrt - the sqrt function from the C standard library is much
# faster
from libc.math cimport sqrt

# disable checks that ensure that array indices don't go out of bounds. this is
# faster, but you'll get a segfault if you mess up your indexing.
@cython.boundscheck(False)
# this disables 'wraparound' indexing from the end of the array using negative
# indices.
@cython.wraparound(False)
def euclidean(double [:, :] A):

    # declare C types for as many of our variables as possible. note that we
    # don't necessarily need to assign a value to them at declaration time.
    cdef:
        # Py_ssize_t is just a special platform-specific type for indices
        Py_ssize_t nrow = A.shape[0]
        Py_ssize_t ncol = A.shape[1]
        Py_ssize_t ii, jj, kk

        # this line is particularly expensive, since creating a numpy array
        # involves unavoidable Python API overhead
        np.ndarray[np.float64_t, ndim=2] D = np.zeros((nrow, nrow), np.double)

        double tmpss, diff

    # another advantage of using Cython rather than broadcasting is that we can
    # exploit the symmetry of D by only looping over its upper triangle
    for ii in range(nrow):
        for jj in range(ii + 1, nrow):
            # we use tmpss to accumulate the SSD over each pair of rows
            tmpss = 0
            for kk in range(ncol):
                diff = A[ii, kk] - A[jj, kk]
                tmpss += diff * diff
            tmpss = sqrt(tmpss)
            D[ii, jj] = tmpss
            D[jj, ii] = tmpss  # because D is symmetric

    return D


from setuptools import setup, Extension
from codecs import open
import sys
import os
import numpy


def is_platform_mac():
    return sys.platform == 'darwin'


def is_platform_windows():
    return sys.platform == 'win32' or sys.platform == 'cygwin'


try:
    from Cython.Distutils import build_ext
except ImportError:
    use_cython = False
else:
    use_cython = True

if use_cython:
    sourcefiles = ['cscipy/distance.pyx']
else:
    sourcefiles = ['cscipy/distance.cpp']

# Fix compatibility when compiling on Mac Mojave.
# Explanation: https://github.com/pandas-dev/pandas/issues/23424#issuecomment-446393981
# Code credit: https://github.com/pandas-dev/pandas/pull/24274/commits/256faf2011a12424e684a42c147e1ba7ac32c6fb
if is_platform_mac():
    import _osx_support
    import distutils.sysconfig

    if not 'MACOSX_DEPLOYMENT_TARGET' in os.environ:
        current_system = list(map(int, _osx_support._get_system_version().split('.')))
        python_osx_target_str = distutils.sysconfig.get_config_var('MACOSX_DEPLOYMENT_TARGET')
        python_osx_target = list(map(int, python_osx_target_str.split('.')))
        if python_osx_target < [10, 9] <= current_system:
            os.environ['MACOSX_DEPLOYMENT_TARGET'] = '10.9'

if is_platform_windows():
    extra_compiler_args = []
else:
    extra_compiler_args = [
        '-std=c++11',
        '-Wno-sign-compare',
        '-Wno-incompatible-pointer-types',
        '-Wno-unused-variable',
        '-Wno-absolute-value',
        '-Wno-visibility',
        '-Wno-#warnings',
    ]

ext_modules = [
    Extension('cscipy.distance',
              sourcefiles,
              language='c++',
              include_dirs=[numpy.get_include()],
              extra_compile_args=extra_compiler_args,
              ),
]

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup_args = dict(
    name='cscipy',
    ext_modules=ext_modules,
    license='MIT',
    packages=['cscipy'],
    version='0.0.1',
    author=['Yukio Fukuzawa'],
    description='Some scipy functions rewritten in Cython',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/fzyukio/cscipy',
    keywords=['scipy', 'cython', 'distance'],
    install_requires=['Cython', 'numpy'],
)

if use_cython:
    setup_args['cmdclass'] = {'build_ext': build_ext}

setup(
    **setup_args
)
